"""
SpiffyPlots base module.
This module contains class definitions that often inherit matplotlib's classes to
add spiffy functionality while still allowing for the original matplotlib syntax.
"""

from ._figure import Figure
from ._panel import Panel
