"""Top-level package for SpiffyPlots."""

__author__ = """Julian Rossbroich"""
__email__ = "julian.rossbroich@fmi.ch"
__version__ = "0.6.0"

from .multipanel import MultiPanel
from .lineplots import multiline
